<?php

namespace Drupal\tmgmt_asymmetric_block\EventSubscriber;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\TempStore\SharedTempStoreFactory;
use Drupal\layout_builder\LayoutEntityHelperTrait;
use Drupal\layout_builder\Field\LayoutSectionItemList;
use Drupal\layout_builder\Plugin\SectionStorage\OverridesSectionStorage;
use Drupal\layout_builder\SectionComponent;
use Drupal\layout_builder\SectionStorageInterface;
use Drupal\tmgmt_asymmetric_block\Events\TmgmtJobItemEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Drupal\layout_builder\Section;
use Drupal\Component\Uuid\UuidInterface;

/**
 * Class UserLoginSubscriber.
 *
 * @package Drupal\tmgmt_asymmetric_block\EventSubscriber
 */
class TmgmtJobItemSubscriber implements EventSubscriberInterface {

  use LayoutEntityHelperTrait;

  /**
   * The user interface.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The logger channel.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The job item entity interface.
   *
   * @var \Drupal\tmgmt\JobItemInterface
   */
  protected $jobItem;

  /**
   * The Job entity interface.
   *
   * @var \Drupal\tmgmt\JobInterface
   */
  protected $job;

  /**
   * The target language code id.
   *
   * @var string
   */
  protected $targetLangcode;

  /**
   * The temp store repository interface.
   *
   * @var \Drupal\Core\TempStore\SharedTempStoreFactory
   */
  protected $sharedTempStore;

  /**
   * The Uuid interface.
   *
   * @var \Drupal\Component\Uuid\UuidInterface
   */
  protected $uuidService;

  /**
   * Initializing services.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    LoggerChannelFactoryInterface $logger_factor,
    SharedTempStoreFactory $shared_temp_store,
    UuidInterface $uuid_service
    ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->logger = $logger_factor->get('tmgmt');
    $this->sharedTempStore = $shared_temp_store;
    $this->uuidService = $uuid_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [
      // Static class constant => method on this class.
      TmgmtJobItemEvent::ACCEPTED => ['onAccepted'],
    ];
  }

  /**
   * Assigned the job, job item and target language code to property.
   *
   * @param \Drupal\tmgmt_asymmetric_block\Events\TmgmtJobItemEvent $event
   *   The job item interface.
   */
  public function initialized(TmgmtJobItemEvent $event):void {
    // Get the current job item from the presave entity.
    $this->jobItem = $event->getJobItem();
    // The job that the job items belongs to.
    $this->job = $this->jobItem->getJob();
    // The target language object to translate to.
    $this->targetLangcode = $this->job->getTargetLanguage()->getId();
  }

  /**
   * Handles the presave entity when a job item is accepted.
   *
   * @param \Drupal\tmgmt_asymmetric_block\Events\TmgmtJobItemEvent $event
   *   The job item interface.
   */
  public function onAccepted(TmgmtJobItemEvent $event) {
    $this->initialized($event);
    /** @var \Drupal\Core\Entity\ContentEntityInterface|false */
    $entity = $this->getEntityJobItem();
    if (!$entity) {
      return;
    }
    $this->handlePreSave($entity);
  }

  /**
   * Handles saving of the layout and its component.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The node interface.
   */
  public function handlePreSave(ContentEntityInterface $entity) {
    // Check to see if layout builder is enabled for this $entity.
    if (!$this->isLayoutCompatibleEntity($entity)) {
      return;
    }

    // Get the source node.
    $this->setSourceEntity($entity);
    $translated_entity = $this->getTranslatedEntity($entity);

    // Check if the job item entity ID is different from the one we are saving
    // and current entity type is the same one as the job item type.
    if ($this->jobItem->getItemType() === $entity->getEntityType()->id() && $this->jobItem->getItemId() == $entity->id()) {
      return;
    }

    $this->updateSectionComponents($translated_entity);
  }

  /**
   * Set the source entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The content entity interface.
   */
  public function setSourceEntity(ContentEntityInterface $entity) {
    $this->sourceEntity = $entity;
  }

  /**
   * Get the source entity.
   */
  public function getSourceEntity() {
    return $this->sourceEntity;
  }

  /**
   * Update the each components in the section for layout builder translation.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $translated_entity
   *   The translated entity interface.
   *
   *   Save the entity event if no block has been updated, mainly use on
   *   updating the main node entity.  Defaults is FALSE.
   */
  public function updateSectionComponents(ContentEntityInterface $translated_entity) {
    $blockUpdated = FALSE;
    $targetLangcode = $this->job->getTargetLangcode();
    $translated_entity
      ->get(OverridesSectionStorage::FIELD_NAME)
      ->setLangcode($targetLangcode);

    $this->entity = $translated_entity;
    /** @var \Drupal\layout_builder\Field\LayoutSectionItemList|null $translated_layout_field */
    $translated_layout_field = $translated_entity->get(OverridesSectionStorage::FIELD_NAME);
    if (!$translated_layout_field) {
      return;
    }
    if ($translated_layout_field->isEmpty()) {
      $this->cloneSection($translated_entity);
    }

    $entity_type = $this->jobItem->getItemType();
    /** @var \Drupal\layout_builder\SectionStorageInterface $section_storage */
    $section_storage = $this->getSectionStorageForEntity($translated_entity);
    if ($section_storage) {
      /** @var \Drupal\layout_builder\Section $section */
      foreach ($this->getSourceEntity()->get(OverridesSectionStorage::FIELD_NAME)->getSections() as $delta => $section) {
        foreach ($section->getComponents() as $component) {
          $cloned_component = clone $component;
          $section_component_config = $cloned_component->get('configuration');
          if ($this->canUpdateComponent($section_component_config, $entity_type)) {
            $blockUpdated = $this->updateComponent($section_storage, $cloned_component, $translated_layout_field, $delta);
            if ($blockUpdated) {
              // Exit out of the foreach completely if we find a match.
              break 2;
            }
          }
        }
      }
    }

    if ($blockUpdated) {
      $sections = !$translated_layout_field->isEmpty()
        ? $translated_layout_field->getSections()
        : [];

      $translated_entity->set(OverridesSectionStorage::FIELD_NAME, $sections);
      $translated_entity->setNewRevision();
      $translated_entity->save();
      // Since content in layout is being updated in code, there will not be
      // a revert option, we want to cleanup any stale tempstore for that node.
      $collection = 'layout_builder.section_storage.overrides';
      $key = implode('.', [
        $translated_entity->getEntityTypeId(),
        $translated_entity->id(),
        // The view mode to clear.
        'default',
        $targetLangcode,
      ]);
      $tempstore = $this->sharedTempStore->get($collection, $key);
      if ($tempstore) {
        $tempstore->delete($key);
      }
    }
  }

  /**
   * Clone the layout builder section with new uuid for section component.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $translated_entity
   *   The translated entity.
   */
  public function cloneSection(ContentEntityInterface $translated_entity) {
    $cloned_sections = [];
    foreach ($this->getSourceEntity()->get(OverridesSectionStorage::FIELD_NAME)->getSections() as $delta => $section) {
      $cloned_sections[$delta] = clone $section;
      // Replace all the clone uuid with new ones.
      foreach ($cloned_sections[$delta]->getComponents() as $component) {
        $cloned_sections[$delta]->removeComponent($component->getUuid());
      }
    }

    $translated_entity->set(OverridesSectionStorage::FIELD_NAME, $cloned_sections);
  }

  /**
   * Allow to update the component.
   *
   * @param array $section_component_config
   *   The section configuration component.
   * @param string $entity_type
   *   The entity type of the block component.
   *
   * @return bool
   *   Returns true for the block component.
   */
  public function canUpdateComponent(array $section_component_config, $entity_type) {
    // Only update the component inline blocks.
    return !empty($section_component_config['block_revision_id']) && $entity_type === 'block_content';
  }

  /**
   * Set the value of the component before it is updated.
   *
   * @param \Drupal\layout_builder\SectionStorageInterface $section_storage
   *   The section storage interface.
   * @param \Drupal\layout_builder\SectionComponent $component
   *   The component in the layout section.
   * @param \Drupal\layout_builder\Field\LayoutSectionItemList $translated_layout_field
   *   The translation override section list.
   * @param int $delta
   *   The delta of the source section.
   */
  public function updateComponent(
    SectionStorageInterface $section_storage,
    SectionComponent $component,
    LayoutSectionItemList $translated_layout_field,
    $delta
  ): bool {
    /** @var \Drupal\Core\Entity\ContentEntityStorageInterface $block_storage */
    $block_storage = $this->entityTypeManager->getStorage($this->jobItem->getItemType());

    // Load current block and get the block ID.
    /** @var \Drupal\layout_builder\Plugin\Block\InlineBlock $plugin */
    $plugin = $component->getPlugin();
    $configuration = $plugin->getConfiguration();
    $data = $this->jobItem->getData();

    if (!empty($data['info'][0]['value']['#translation']['#entity_created'])) {
      $entity_created = $data['info'][0]['value']['#translation']['#entity_created'];
      $translated_block_revision = $block_storage->loadRevision($entity_created['#revision_id']);
      /** @var \Drupal\block_content\Entity\BlockContent $original_block */
      $original_block = $block_storage->loadRevision($configuration['block_revision_id']);
      // Only continue when the job item block id and the component block ID
      // match.
      if ($translated_block_revision && $original_block->id() == $this->jobItem->getItemId()) {
        /** @var \Drupal\block_content\BlockContentInterface $block_content_revision */
        \Drupal::service('inline_block.usage')->addUsage($translated_block_revision->id(), $this->getEntity());
        $block_content_revision = $translated_block_revision;

        $configuration['block_revision_id'] = (string) $entity_created['#revision_id'];
        $configuration['label'] = $block_content_revision->label();
        $section = $translated_layout_field->getSection($delta);
        $component->setConfiguration($configuration);
        $this->addComponentToSection($section, $component);
        return TRUE;
      }
    }

    return FALSE;
  }

  /**
   * Add component to section.
   *
   * @param \Drupal\layout_builder\Section $section
   *   A section of the layout.
   * @param \Drupal\layout_builder\SectionComponent $component
   *   The section component.
   */
  public function addComponentToSection(Section $section, SectionComponent $component) {
    $uuid = $this->uuidService->generate();
    $component->set('uuid', $uuid);
    if ($component->getWeight() > count($section->getComponents())) {
      $section->appendComponent($component);
    }
    else {
      $section->insertComponent($component->getWeight(), $component);
    }
  }

  /**
   * Gets the primary entity from the job.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface|false
   *   Returns the untranslated entity or false if the entity is not found.
   */
  public function getEntityJobItem() {
    $entity_job_items = $this->entityTypeManager->getStorage('tmgmt_job_item')->loadByProperties([
      'tjid' => $this->jobItem->getJobId(),
    ]);
    if (!empty($entity_job_items)) {
      // Get the object from the first array key..
      /** @var \Drupal\tmgmt\JobItemInterface $entity_job_item */
      $entity_job_item = reset($entity_job_items);
      $entity = $this->entityTypeManager->getStorage($entity_job_item->getItemType())->load($entity_job_item->getItemId());
      if ($entity) {
        return $entity;
      }
    }
    return FALSE;
  }

  /**
   * Get the translated entity content.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The untranslated entity interface.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface
   *   The translated entity.
   */
  public function getTranslatedEntity(ContentEntityInterface $entity) {
    $langcode = $this->targetLangcode;
    if ($entity->isTranslatable() && !$entity->hasTranslation($langcode)) {
      // Add a new translation if none existed.
      $title = $entity->label();
      $translation = $entity->addTranslation($langcode, $entity->toArray());
      // Set any required field on the entity before saving.
      $label_key = $entity->getEntityType()->getKeys()['label'];
      $translation->set($label_key, $title);
    }
    else {
      $translation = $entity->getTranslation($langcode);
    }
    return $translation;
  }

  /**
   * Get the translated entity.
   */
  public function getEntity() {
    return $this->entity;
  }

}
